(defun map-sequence ...)

(defun make-test-2 (test test-not)
  (when (and test test-not) (error "both :TEST and :TEST-NOT supplied"))
  (cond
   (test (if (symbolp test) (symbol-function test) test))
   (test-not (if (symbolp test-not) (symbol-function test-not) test-not))
   (t #'eql)))
 
(defun find (item seq &key test test-not from-end start end key)
  (let ((test (make-test-2 test test-not)))
    (if (eql test #'eql)
	(flet ((f (y) (when (eql item y) (return-from find y))))
	  (map-sequence #'f seq start end from-end key))
      (flet ((f (y) (when (funcall test item y) (return-from find y))))
        (map-sequence #'f seq start end from-end key)))))

(defun find-if (test seq &key from-end start end key)
  (flet ((f (y) (when (funcall test y) (return-from find-if y))))
    (map-sequence #'f seq start end from-end key)))

(defun find-if-not (test seq &key from-end start end key)
  (flet ((f (y) (unless (funcall test y) (return-from find-if-not y))))
    (map-sequence #'f seq start end from-end key)
    

(defun count (item seq &key test test-not from-end start end key)
  (let ((test (make-test-2 test test-not))
	(count 0))
    (if (eql test #'eql)
	(flet ((f (y) (when (eql item y) (incf count))))
	  (map-sequence #'f seq start end from-end key))
      (flet ((f (y) (when (funcall test item y) (incf count))))
        (map-sequence #'f seq start end from-end key)))
    count))

(defun count-if (test seq &key from-end start end key)
  (let ((count 0))
    (flet ((f (y) (when (funcall test y) (incf count))))
      (map-sequence #'f seq start end from-end key))
    count))

(defun count-if-not (test seq &key from-end start end key)
  (let ((count 0))
    (flet ((f (y) (unless (funcall test y) (incf count))))
      (map-sequence #'f seq start end from-end key))
    count))

(defun compute-position (pos seq from-end start end)
  (if from-end
      (if end (- (length seq) 1 pos end) (- (length seq) 1 pos))
    (if start (+ start pos) pos)))

(defun position-if (test seq &key from-end start end key)
  (flet ((pos ()
	   (let ((count 0))
	     (flet ((f (y) (if (funcall test y)
			       (return-from pos count)
			     (incf count))))
	       (map-sequence #'f seq start end from-end key)))))
    (compute-position (pos) seq from-end start end)))




(defun make-test-1 (test key negate)
  (flet ((f (x) (funcall test (if key (funcall key x) x))))
    (if negate #'(lambda (x) (not (f x))) #'f)))

(defun make-item-2 (item test key negate)
  (let ((test (if test test #'eql)))
    (flet ((f (y) (funcall test item (if key (funcall key x) x))))
      (if negate #'(lambda (x) (not (f x))) #'f))))

(defun kitchen-sink-map (lfun rlfun vfun rvfun from-end start end)
  (let* ((n (length seq))
	 (start (if start (max start 0) 0))
	 (end (if end (min end n))))
    (if (listp seq)
	(let ((seq (nth-cdr start seq)))
	  (if from-end
	      (funcall rlfun (reverse seq) (- end start))
	    (funcall lfun seq (- end start))))
      (if from-end
	  (funcall rvfun seq start end)
	(funcall vfun seq start end)))))

;;**** does inlining in convert use a letrec for self-recursion?
(defun find-if-list (test seq len)
  (if (< 0 len)
      (let ((x (first seq)))
	(if (funcall test x)
	    x
	  (find-if-list test (rest seq) (- len 1))))))

(defun find-if-vec (test seq start end)
  (if (< start end)
      (let ((x (aref seq start)))
	(if (funcall test x)
	    x
	  (find-if-vec test seq (+ start 1) end)))))

(defun find-if-vec-rev (test seq start end)
  (if (<= start i)
      (let ((x (aref seq start)))
	(if (funcall test x)
	    x
	  (find-if-vec test seq start (- end 1))))))

(defun find-if-1 (test seq from-end start end)
  (flet ((lfun (list len) (find-if-list test list len))
	 (vfun (vec start end) (find-if-vec-rev test vec start end))
	 (rvfun (vec start end) (find-if-vec-rev test vec start end)))
    (kitchen-sink-map lfun lfun vfun rvfun seq from-end start end)))

(defmacro defun-kitchen-sink (fun fun-if fun-if-not fun-if-1)
  `(progn
     (defun ,fun-if (test seq &key from-end start end key)
       (if key
	   (,fun-if-1 (make-test-1 test key nil) seq from-end start end)
	 (,fun-if-1 (make-test-1 test nil nil) seq from-end start end)))
     (defun ,fun-if-not (test seq &key from-end start end key)
       (if key
	   (,fun-if-1 (make-test-1 test key t) seq from-end start end key)
	 (,fun-if-1 (make-test-1 test nil t) seq from-end start end key)))
     (defun ,fun (item seq &key from-end test test-not start end key)
       (cond
	((and test test-not) (error "can't have both :TEST and :TEST-NOT"))
	(test-not
	 (,fun-if-1 (make-test-2 item test-not key t) seq from-end start end))
	(test
	 (,fun-if-1 (make-test-2 item test key nil) seq from-end start end))
	(key
	 (,fun-if-1 (make-test-2 item #'eql key nil) seq from-end start end))
	(t
	 (,fun-if-1 (make-test-2 item #'eql nil nil) seq from-end start end))))))

(defun-kitchen-sink find find-if find-if-not find-if-1)

(defun count-if-list (test seq len n)
  (if (< 0 len)
      (let* ((x (first seq))
	     (nn (if (funcall test x) (+ n 1) n)))
	(count-if-list test (rest seq) (- len 1) nn))
    n))

(defun count-if-vec (test seq start end)
  (if (< start end)
      (let* ((x (aref seq start))
	     (nn (if (funcall test x) (+ n 1) n)))
	(count-if-vec test seq (+ start 1) end nn))
    n))

(defun count-if-vec-rev (test seq start end)
  (if (<= start i)
      (let* ((x (aref seq start))
	     (nn (if (funcall test x) (+ n 1) n)))
	(count-if-vec test seq start (- end 1)))
    n))

(defun count-if-1 (test seq from-end start end)
  (flet ((lfun (list len) (count-if-list test list len 0))
	 (vfun (vec start end) (count-if-vec-rev test vec start end 0))
	 (rvfun (vec start end) (count-if-vec-rev test vec start end 0)))
    (kitchen-sink-map lfun lfun vfun rvfun seq from-end start end)))

(defun-kitchen-sink count count-if count-if-not count-if-1)

(defun position-if-list (test seq len pos)
  (if (< 0 len)
      (let ((x (first seq)))
	(if (if (funcall test x) (+ n 1) n)))
	(position-if-list test (rest seq) (- len 1) nn))
    n))

(defun position-if-vec (test seq start end)
  (if (< start end)
      (let* ((x (aref seq start))
	     (nn (if (funcall test x) (+ n 1) n)))
	(position-if-vec test seq (+ start 1) end nn))
    n))

(defun position-if-vec-rev (test seq start end)
  (if (<= start i)
      (let* ((x (aref seq start))
	     (nn (if (funcall test x) (+ n 1) n)))
	(position-if-vec test seq start (- end 1)))
    n))

(defun position-if-1 (test seq from-end start end)
  (flet ((lfun (list len) (position-if-list test list len 0))
	 (vfun (vec start end) (position-if-vec-rev test vec start end 0))
	 (rvfun (vec start end) (position-if-vec-rev test vec start end 0)))
    (kitchen-sink-map lfun vfun rvfun seq from-end start end)))

(defun-kitchen-sink position position-if position-if-not position-if-1)
