;; random state for the simulation
;; the with-input-from-string nonsense is needed because xlisp doesn't
;; allow the #$ combination.
;;
;; For some bizzare reason resetting the random state in kcl doesn't
;; seem to be working right.

(setf *my-random-state*
      #+xlisp '#( 716126655  562082643 1548792663 1457602827 1238184623
                 2051834243  138939847 1620054715 1340574367  764002995
                  960913207   33230699  721498511 2034804451  314578855
                 1057351451  605820799  925654547 2147483647      65536
                          5         17)
      #+kcl (with-input-from-string (f "#$679108871") (read f)))
;;
;; Find patches using Rem's algorithm for finding equivalence
;; classes (in A Dicipline of Programming -- from Charlie Geyer)
;;

(defmacro while (c &rest b) `(loop (if (not ,c) (return)) ,@b))

(defun find-patches (edges bonds f)
  (let ((nvert (length f))
	(nedge (length edges)))
;    (declare (fixnum nvert nedge))

    (dotimes (v nvert)
;      (declare (fixnum v))
      (setf (aref f v) v))

    (dotimes (e nedge)
;      (declare (fixnum e))
      (let ((edge (aref edges e))
	    (has-bond (aref bonds e)))
	(if has-bond
	  (let ((p0 (first edge))
		(q0 (second edge)))
	    (let ((p1 (aref f p0))
		  (q1 (aref f q0)))
;	      (declare (fixnum p0 q0 p1 q1))
	      (while (/= p1 q1)
		(if (< q1 p1)
		    (progn
		      (setf (aref f p0) q1)
		      (setf p0 p1)
		      (setf p1 (aref f p1)))
		    (progn
		      (setf (aref f q0) p1)
		      (setf q0 q1)
		      (setf q1 (aref f q1))))))))))

    (dotimes (v nvert)
;      (declare (fixnum v))
      (setf (aref f v) (aref f (aref f v))))

    (let ((current -1)
	  (count 0))
;      (declare (fixnum current count))
      (dotimes (v nvert count)
;        (declare (fixnum v))
        (let ((c (aref f v)))
	  (if (< current c)
	      (progn (setf current c)
		     (setf count (+ count 1)))))))))

(defun make-grid-edges (n)
  (let ((g (make-array (* 2 n (- n 1))))
	(k 0))
    (flet ((add-edge (a b)
	     (setf (aref g k) (list a b))
	     (setf k (+ k 1)))
	   (point (i j) (+ (* i n) j)))
      (dotimes (i n)
        (dotimes (j (- n 1))
	  (add-edge (point i j) (point i (+ j 1)))
	  (add-edge (point j i) (point (+ j 1) i)))))
    g))

(defun randomize-bonds (b p)
  (dotimes (i (length b))
    (setf (aref b i) (if (< (random 1.0) p) t nil))))

(defun setup-grid (n)
  (setf gr (make-grid-edges n))
  (setf b (make-array (length gr)))
  (setf f (make-array (* n n)))
  (setf b1 (copy-array b))
  (setf f1 (copy-array f))
  nil)

(defun color-vertices (f)
  (let* ((nvert (length f))
	 (colors '(black white))
	 (ncols (length colors))
	 (same-color t)
	 (first-color nil))

    (let ((current -1))
      (dotimes (v nvert)
        (let ((c (aref f v)))
	  (when (< current c)
		(let ((new-color (select colors (random ncols))))
		  (unless first-color (setf first-color new-color))
		  (unless (eql new-color first-color) (setf same-color nil))
		  (setf (aref f v) new-color)))
	  (setf current c))))
    
    (dotimes (v nvert)
      (let ((c (aref f v)))
	(if (numberp c)
	    (setf (aref f v) (aref f (aref f v))))))
    same-color))

(defun make-bonds (edges b f p)
  (dotimes (e (length b))
    (let* ((edge (aref edges e))
	   (i (first edge))
	   (j (second edge)))
      (setf (aref b e)
	    (if (and (eql (aref f i) (aref f j))
		     (< (random 1.0) p))
		t
	      nil)))))

(defun sw (n p)
  (setf *random-state* (make-random-state *my-random-state*))
  (let ((pc (make-array n))
	(sc (make-array n)))
    (dotimes (i n (list pc sc))
      (make-bonds gr b f p)
      (setf (aref pc i) (find-patches gr b f))
      (setf (aref sc i) (color-vertices f)))))

(defun swa (n p)
  (setf *random-state* (make-random-state *my-random-state*))
  (let ((pc (make-array (* 2 (floor (/ n 2)))))
	(sc (make-array (* 2 (floor (/ n 2)))))
	(ncols 2))
    (dotimes (i (floor (/ n 2)) (list pc sc))
      (make-bonds gr b f p)
      (setf (aref pc (* 2 i)) (find-patches gr b f))
      (setf (aref sc (* 2 i)) (color-vertices f))
      ;; independence step
      (randomize-bonds b1 p)
      (let* ((cx (aref pc (* 2 i)))
	     (cy (find-patches gr b1 f1))
	     (lwx (* (log ncols) cx))
	     (lwy (* (log ncols) cy)))
	(format t "~a~%" (list lwx lwy))
	(when (< (log (random 1.0)) (- lwy lwx))
	      (dotimes (j (length b)) (setf (aref b j) (aref b1 j)))
	      (dotimes (j (length f)) (setf (aref f j) (aref f1 j)))
	      (color-vertices f)
	      (setf (aref pc (+ (* 2 i) 1)) cy))))))

(defun f (p p1)
  (randomize-bonds b1 p1)
  (let* ((c1 (select (sw 1 p) 0))
	 (c2 (find-patches gr b1 f1))
	 (n (length b))
	 (k1 (count t b))
	 (k2 (count t b1))
	 (ncols 2))
    (list (- (+ (* k2 (- (log p) (log p1)))
		(* (- n k2) (- (log (- 1 p)) (log (- 1 p1))))
		(* (log ncols) c2))
	     (+ (* k1 (- (log p) (log p1)))
		(* (- n k1) (- (log (- 1 p)) (log (- 1 p1))))
		(* (log ncols) c1)))
	  c1 c2)))

(defun h (n p)
  (let ((result nil))
    (dotimes (i n result)
      (setf result (cons (g p) result)))))
(defun g (p)
  (randomize-bonds b1 p)
  (let ((c (find-patches gr b1 f1)))
    (* (log 2) c)))
(defun k (c x)
  (let* ((y (exp (- x (mean x))))
	 (w (/ y (mean y))))
    (* c (^ (mean (pmin 1 (/ w c))) 2))))

#|
Try a mixture of binomials (won't work!)
Only try restarts if some condition holds (is this legitimate?)
|#

(defun pi-s-eq (lc log-w)
  (let* ((lw (- log-w))
	 (lwm (mean lw))
	 (lmean (+ lwm (log (mean (exp (- lw lwm)))))))
    (^ (mean (exp (+ (* 0.5 (+ lc lmean)) (- (pmin lw (- lc)) lmean)))) 2)))

(defun independence-r (cx cy lc)
  (if (and cx cy)
      (let ((wx (exp (- (* (log 2) cx) lc)))
	    (wy (exp (- (* (log 2) cy) lc))))
	(min 1 (/ (min wx wy)) (max wx wy)))
      0))

(defun ecdf (x)
  (let* ((x (coerce (sort-data x) 'list))
	 (n (length x))
	 (xr nil)
	 (cr nil)
	 (current (first x))
	 (count 0))
    (dolist (v x)
      (when (/= current v) 
	    (setf xr (cons current xr))
	    (setf cr (cons (/ count n) cr))
	    (setf current v))
      (setf count (+ count 1)))
    (list (reverse (cons current xr)) (reverse (cons (/ count n) cr)))))

(defun pp-geometric (x k)
  (let ((e (ecdf x))
	(p (/ k (mean x))))
    (list (second e)
	  (- 1 (^ (- 1 p) (/ (first e) k))))))

;; Grid of 32 x 32, p = 0.8
;; For pure Gibbs with split on all one color, using run of 5000 gives
;; 197 regenerations, or 196 complete tours; this gives an estimated rate
;; of 0.039.
;;
;; Optimal log c for independence steps is about 4; estimated regeneration
;; rate for independence steps is 0.166.
;; Based on run of 5000, estimated rate is 0.073; 358 regenerations, or 357
;; complete tours, are observed.

#|
;; generate the coin flips
(def s (if-else (second x) 1 0))

;; compute the tour lengths
(def n (difference (which (= 1 s))))

;; compute the pp-plot
(def pp-n (pp-geometric n 2))

(setf c-opt 4)

;; compute r's
(def ra (if-else (evenp (iseq (length (first xa))))
		  (maplist #'(lambda (x)
			       (if (consp (cdr x)) 
				   (independence-r (first x) (second x) c-opt) 
				   0))
			   (first xa))
		  0))

;; estimate regeneration rate using $r(x,y)$
(mean ra)

;; generate the coin flips
(def sa (if-else (< (uniform-rand (length ra)) ra) 1 0))

;; compute the tour lengths
(def na (difference (which (= 1 sa))))

;; compute the pp-plot
(def pp-na (pp-geometric na 2))
|#

(defun count-concordant-pairs (gr f)
  (reduce #'(lambda (x y)
	      (if (eq (select f (first y)) (select f (second y))) (+ x 1) x))
	  gr
	  :initial-value 0))

(defun count-colors (f)
  (let ((nb 0)
	(nw 0))
    (dotimes (i (length f) (list nw nb))
      (if (eq (aref f i) 'white) (setf nw (+ nw 1)) (setf nb (+ nb 1))))))

(defun count-bonds (b)
  (let ((nb 0)
	(nn 0))
    (dotimes (i (length b) (list nb nn))
      (if (aref b i) (setf nb (+ nb 1)) (setf nn (+ nn 1))))))

(defun log-bcoef (n m)
  (- (log-gamma (+ n 1))
     (log-gamma (+ m 1))
     (log-gamma (+ 1 (- n m)))))

(defun log-w-1 (gr f beta p)
  (let ((conpairs (count-concordant-pairs gr f))
	(cols (count-colors f)))
    (- (* beta conpairs)
       (sum (* cols (log (list p (- 1 p))))))))

(defun randomize-colors (f p)
  (dotimes (i (length f))
    (setf (aref f i) (if (< (random 1.0) p) 'white 'black))))

(defun swa1 (n p p1)
;  (setf *random-state* (make-random-state *my-random-state*))
  (let ((pc (make-array (* 2 (floor (/ n 2)))))
	(sc (make-array (* 2 (floor (/ n 2)))))
	(beta (- (log (- 1 p)))))
    (dotimes (i (floor (/ n 2)) (list pc sc))
      (make-bonds gr b f p)
      (find-patches gr b f)
      (setf (aref sc (* 2 i)) (color-vertices f))
      (setf (aref pc (* 2 i)) (log-w-1 gr f beta p1))
      ;; independence step
      (randomize-colors f1 p1)
      (let* ((lwx (aref pc (* 2 i)))
	     (lwy (log-w-1 gr f1 beta p1)))
	(when (< (log (random 1.0)) (- lwy lwx))
	      (dotimes (j (length b)) (setf (aref b j) (aref b1 j)))
	      (dotimes (j (length f)) (setf (aref f j) (aref f1 j)))
	      (find-patches gr b f p)
	      (setf (aref pc (+ (* 2 i) 1)) lwy))))))

#|
(setf p (send graph-window-proto :new))

(defmeth p :redraw ()
    (let* ((n 32)
	   (width (send self :canvas-width))
	   (height (send self :canvas-height))
	   (cw (round (/ width n)))
	   (ch (round (/ height n))))
      (let ((k 0))
	(dotimes (i 32)
		 (dotimes (j 32)
			  (send self :draw-color (aref f k))
			  (send self :paint-rect (* i cw) (* j ch) cw ch)
			  (setf k (+ k 1)))))))

(send p :menu (send menu-proto :new "SW"))
(send (send p :menu) :append-items
      (send menu-item-proto :new "Step" :action #'(lambda () (swp *sw-p*))))
(setf *sw-p* .8)
|#

(defun log-w-2 (gr f beta p)
  (- (* beta (count-concordant-pairs gr f))
     (sum (* (count-colors f) (log (* p 1024))))))

;; Try some scheme with local dependence (every other point indep, fill by
;; conditionals somehow. Or checkerboard pattern -- white cells independent,
;; red ones by conditionals.
