(setf g '(lambda (e compile-time-too env stream)
  (setq e (expand-one e env))
  (case (first e)
        (progn
          (dolist (e (rest e)) (compile-one e compile-time-too env stream)))
        (defun
         (let* ((name (second e))
                (db (xlisp::split-declarations (rest (rest (rest e)))))
                (decls (first db))
                (b (second db))
                (doc (third db))
                (fexpr `(labels ((,name ,(third e) ,@decls ,@b)) #',name))
                (cexpr (compile-back (optimize (compile-front-form fexpr env))))
                (cform (if doc
                           `(progn
                              (xlisp::%set-get ',name 'function-documentation ,doc)
                              (xlisp::install-function ',name ,cexpr))
                         `(xlisp::install-function ',name ,cexpr))))
           (accumulate-compiled-form cform stream)))
        (macrolet
         (flet ((mkmacro (x)
                         (let ((name (first x))
                               (args (second x))
                               (body (rest (rest x))))
                           (list name (parse-macro name args body env)))))
               (multiple-value-bind (decls body)
                                    (split-decls-body (rest (rest e)) nil)
                 (let* ((macs (mapcar #'mkmacro (second e)))
                        (nenv (augment-environment env :macro macs))
                        (nnenv (augment-environment nenv :declare decls)))
                   (compile-one body compile-time-too nnenv stream)))))
        (eval-when
         (let* ((sits (second e))
                (body `(progn ,@(rest (rest e))))
                (lt (or (member 'load sits) (member :load-toplevel sits)))
                (ct (or (member 'compile sits)
                        (member :compile-toplevel sits)))
                (ex (or (member 'eval sits) (member :execute sits))))
           (cond
            ((or (and lt ct)
                 (and lt (not ct) ex compile-time-too))
             (compile-one body t env stream))
            ((or (and lt (not ct) (not compile-time-too))
                 (and lt (not ct) (not ex)))
             (compile-one body nil env stream))
            ((or (and (not lt) ct)
                 (and (not lt) (not ct) ex compile-time-too))
             (eval body)))))
       (t (if compile-time-too (eval e))
          (compile-form e env stream)))))
(setf h '(lambda  (file &key
                          (output-file (merge-pathnames ".fsl" file))
                          (temporary-file (merge-pathnames "cmptmp.fsl" file))
                          (load nil)
                          (print *compile-print*)
                          (verbose *compile-verbose*)
                          ((:print-symbol-package sp) ;;**** do this cleaner
                           *compile-print-symbol-package*))
  (let* ((*package* *package*)
         (*readtable* *readtable*)
         (iname (merge-pathnames ".lsp" file))
         (*compile-file-pathname* iname)
         (*compile-file-truename* (truename iname))
         (conditions nil)
         (eof (list 'eof)))
    (handler-bind ((condition #'(lambda (c) (push c conditions))))
      (with-open-file (in iname)
        (with-open-file (out temporary-file :direction :output)
          (when verbose
                (format t "~&; compiling file ~s~%" iname)
                (force-output))
          (let ((major xlisp::*fsl-major-version*)
                (minor xlisp::*fsl-minor-version*))
            (print-compiled-form `(xlisp::check-fsl-version ,major ,minor)
                                 out sp))
          (with-compiler-environment
            (do ((e (read in nil eof) (read in nil eof)))
                ((eq e eof))
                (when print
                      (let ((*print-level* 2)
                            (*print-length* 3))
                        (format t "~&; compiling ~s ... " e)
                        (force-output)))
                (handler-case
                 (let ((outstuff (list 'progn)))
                   (compile-one e nil xlisp::*global-environment* outstuff)
                   (case (length outstuff)
                         (1 nil)
                         (2 (print-compiled-form (second outstuff) out sp))
                         (t (print-compiled-form outstuff out sp))))
                 (error (c)
                        (format *error-output* "~&Compiler error: ~a~%" c)
                        (print-compiled-form e out sp)))
                (when print
                      (format t "done~%")
                      (force-output))))))
      (rename-file temporary-file output-file)
      (when load (load output-file))
      (values (truename output-file)
              (some #'style-warning-p conditions)
              (not (every #'style-warning-p conditions)))))))
