(in-package "XLISP")

(export 'subtypep)

#|
do need to add eql, member!
don't do eql, member, satisfies, values
don't do compound functions
don't do compound cons's
add structures
add complex
add array
use env
fix up deftypes; add other compound types
streams???
|#

(defparameter *type-table* (make-hash-table :test 'eq))

(defmacro define-type (type args &rest forms)
  (let ((fargs (deftype-arglist-fix args)))
    `(progn (setf (gethash ',type *type-table*) #'(lambda ,fargs ,@forms))
            ',type)))

(defun expand-type (type)
  (let ((name (if (consp type) (first type) type)))
    (if (symbolp name)
	(let ((prop (gethash name *type-table*)))
	  (if prop
	      (expand-type (apply prop (if (consp type) (rest type) nil)))
	    (let ((prop (get name '*type-spec*)))
	      (if prop
		  (expand-type (apply prop (if (consp type) (rest type) nil)))
		type))))
      type)))


(defconstant *base-types*
  '(compiled-function non-compiled-function
    fixnum bignum ratio finite-float non-finite-float
    complex-rational complex-float
    array
    hash-table
    null generic-symbol keyword
    package
    pathname
    random-state
    readtable
    standard-char non-standard-base-char extended-char
    cons))

(defconstant *base-real-types*
  '(fixnum bignum ratio finite-float non-finite-float))

(defconstant *compound-types*
  '(fixnum bignum ratio finite-float non-finite-float
    array
    cons
    compiled-function
    non-compiled-function))

(defconstant *type-operators*
  '(and satisfies eql not values member or))

(defconstant *non-finite-floats*
  (list negative-infinity positive-infinity not-a-number))

(defun base-type-of (x)
  (let ((type (type-of x)))
    (case type
	  (closure 'non-compiled-function)
	  ((byte-code-closure subr fsubr) 'compiled-function)
	  ((fixnum bignum ratio) type)
	  (float (if (member x *non-finite-floats*)
		     'non-finite-float
		   'finite-float))
	  (hash-table type)
	  (symbol (cond
		   ((null x) 'null)
		   ((keywordp x) keyword)
		   (t 'generic-symbol)))
	  (package type)
	  (pathname type) ;;**** doesn't exist yet
	  (random-state type)
	  (readtable type) ;;**** doesn't exist yet
	  (character (if t ;;**** clean this up
			 'standard-char
		       'non-standard-base-char))
	  (extended-char type)
	  (cons type)
	  (t (error "type ~s not handled yet" type)))))

(defun add-type-data (data type vals)
  (let ((e (assoc type data)))
    (if e
        (dolist (v vals data) (pushnew v (rest e)))
       (cons (cons type (remove-duplicates vals)) data))))

(defun range-bound (x)
  (cond
   ((and (consp x) (realp (first x))) (first x))
   ((realp x) x)))

(defun find-type-data (type data)
  (let ((type (expand-type type)))
    (if (consp type)
        (cond
          ((member (first type) '(and or not))
           (find-type-data-list (rest type) data))
          ((member (first type) *base-real-types*)
           (let ((low (range-bound (second type)))
                 (high (range-bound (third type)))
                 (vals nil))
             (when (realp low) (push low vals))
             (when (realp high) (push high vals))
             (if vals
                 (add-type-data data (first type) vals)
                data)))
          ((eql (first type) 'eql)
	   (let ((val (second type)))
	     (add-type-data data (base-type-of val) (list val))))
	  (t data))
       data)))

(defun find-type-data-list (args data)
  (dolist (a args data)
    (setf data (find-type-data a data))))

(defun type-data (x)
  (flet ((make-reps (x)
	   (let* ((x (sort x #'<))
		  (v0 (first x))
		  (low (1- (first x)))
		  (high (1+ (first (last x))))
		  (vals (list low high)))
	     (push v0 vals)
	     (dolist (v (rest x) vals)
	       (push v vals)
	       (push (/ (+ v0 v) 2) vals)
	       (setf v0 v)))))
    (let ((data (find-type-data x nil))
	  (base (length *base-types*))
	  (nd nil))
      (dolist (d data nd)
        (let* ((type (first d))
	       (vals (if (member type *base-real-types*)
			 (make-reps (rest d))
		       (rest d))))
	  (push (list type base vals) nd)
	  (incf base (length vals)))))))

(defun type-and (x y) (if (and x y) (logand x y) nil))
(defun type-or (x y) (if (and x y) (logior x y) nil))
(defun type-not (x) (if x (lognot x) nil))

(defun type-code-reduce (fun types data)
  (reduce fun (mapcar #'(lambda (x) (type-code x data)) types)))
	
(defun type-code (type data)
  (let ((type (expand-type type)))
    (if (consp type)
	(case (first type)
	      ((satisfies cons values) nil) ;**** for now
	      (and (type-code-reduce #'type-and (rest type) data))
	      (or (type-code-reduce #'type-or (rest type) data))
	      (not (type-not (type-code (second type) data)))
	      (eql (let* ((v (second type))
			  (bt (base-type-of v)))
		     (if (member bt *base-real-types*)
			 (type-code `(,bt ,v ,v) data)
		       (eql-type-code bt v data))))
	      (t (if (member (first type) *base-real-types*)
		     (real-type-code type data)
		   nil)))
      (if (member type *base-real-types*)
	  (real-type-code type data)
	(generic-type-code type data)))))

(defun eql-type-code (bt v data)
  (let* ((d (assoc bt data))
	 (base (second d))
	 (vals (third d)))
    (ash 1 (+ base (position v vals)))))

(defun generic-type-code (type data)
  (let ((d (assoc type data))
	(pos (position type *base-types*)))
    (if pos
	(let ((c (ash 1 pos))
	      (base (second d))
	      (vals (third d)))
	  (dolist (v vals c)
	    (setf c (logior c (ash 1 base)))
	    (incf base)))
      nil)))

(defun real-type-code (type data)
  (let* ((tname (if (consp type) (first type) type))
	 (d (assoc tname data)))
    (if d
	(let ((base (second d))
	      (vals (third d))
	      (c 0))
	  (dolist (v vals c)
		  (let ((low (if (consp type) (second type) '*))
			(high (if (consp type) (third type) '*)))
		    (when (in-range v low high)
			  (setf c (logior c (ash 1 base))))
		    (incf base))))
      (ash 1 (position tname *base-real-types*)))))

(defun in-range (v low high)
  (and (cond
	((consp low) (< (first low) v))
	((realp low) (<= low v))
	((eql low '*) t))
       (cond
	((consp high) (< v (first high)))
	((realp high) (<= v high))
	((eql high '*) t))))

;; temporary versions
(define-type real (&rest args)
  (if args
      (let ((low (if (first args) (first args) '*))
            (high (if (second args) (second args) '*)))
        `(or (float ,low ,high) (rational ,low ,high)))
     '(or float rational)))

(define-type float (&rest args)
  (if args
      (let ((low (if (first args) (first args) '*))
            (high (if (second args) (second args) '*)))
        `(finite-float ,low ,high))
     '(or finite-float non-finite-float)))

(define-type rational (&optional low high)
  `(or (integer ,low ,high) (ratio ,low ,high)))

(define-type integer (&optional low high)
  `(or (fixnum ,low ,high) (bignum ,low ,high)))

(define-type t () `(or ,@xlisp::*base-types*))

(define-type symbol () '(or null generic-symbol keyword))

(define-type character ()
  '(or standard-char non-standard-base-char extended-char))

(define-type member (&rest vals)
  `(or ,@(mapcar #'(lambda (x) `(eql ,x)) vals)))

(defun subtypep (x y &optional env)
  (let* ((type `(and ,x (not ,y)))
	 (code (type-code type (type-data type))))
    (cond
     ((eq y t) (values t t))
     (code (values (zerop code) t))
     (t (values nil nil)))))
